function varargout = MPI_Bcast( source, tag, comm, varargin )
% MPI_Bcast  -  broadcast variables to everyone.
%
%   [var1, var2, ...] = ...
%     MPI_Bcast( source, tag, comm, var1, var2, ... )
%
%   Broadcast variables to everyone in comm.
%
%   Sender blocks until all the messages are received,
%   unless MatMMPI_Save_messages(1) has been called.
%

%----------------------------------------------------
% Check the input
%----------------------------------------------------

if nargin < 4
    error('not enough input arguments');
end

if numel(varargin) ~= nargout
    warning(['number of input variables does not match number of '...
        'return values']);
end


%----------------------------------------------------
% Main function
%----------------------------------------------------

% Get processor rank.
my_rank = MPI_Comm_rank(comm);
comm_size = MPI_Comm_size(comm);

% Get the message from your parent, or copy input to output 
% if you're the source;
if my_rank == source
    varargout = varargin;
else 
    varargout = cell(1,nargout);
    [varargout{:}] = MPI_Recv(parent(my_rank, source), tag, comm);

    % for octave
    %cmd = '[';
    %ii = 1;
    %while ii < length(varargin)
    %    cmd = [cmd 'varargout{' num2str(ii) '},'];
    %        ii = ii+1;
    %end
    %cmd = [cmd 'varargout{' num2str(ii) ...
    %       '}] = MPI_Recv(parent(my_rank, source), tag, comm);'];
    %eval(cmd);
end

% Send the message to your left child
dest = left_child(my_rank, source);
if dest < comm_size
    MPI_Send(dest, tag, comm, varargout{:});
end

% Send the message to your right child
dest = right_child(my_rank, source);
if dest < comm_size
    MPI_Send(dest, tag, comm, varargout{:});
end

end % end MPI_Bcast


%----------------------------------------------------
% Helper functions
%----------------------------------------------------

% Tree structure (assuming source == 0): 
%    root = 0, 
%    left child of node i = 2*i+1, 
%    right child of node i = 2*i+2.
%    parent(i) = ceil(ii/2)-1.
% When the source is not 0, we simply swap nodes '0' and 'source' in
% the tree, so 'source' is at the root and '0' is in 'source's
% position.  Otherwise the tree structure remains the same.
%
% The following funcitons simply compute the above relationsihps.

function p = parent(my_rank, source)
    my_rank = swap(my_rank, source);
    p = ceil(my_rank/2)-1;
    p = swap(p, source);
end

function l = left_child(my_rank, source)
    my_rank = swap(my_rank, source);    
    l = 2*my_rank+1;
    l = swap(l, source);
end

function r = right_child(my_rank, source)
    my_rank = swap(my_rank, source);    
    r = 2*my_rank+2;
    r = swap(r, source);
end

function x = swap(x,source)
    if x == source
        x = 0;
    elseif x == 0
        x = source;
    end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% MatlabMPI
% David Ross
% University of Toronto
% dross /at/ cs.toronto.edu
%
% This function is a reimplemntation of a function of the same name by:
% Dr. Jeremy Kepner
% MIT Lincoln Laboratory
% kepner@ll.mit.edu
% (Copyright 2002 Massachusetts Institute of Technology)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright 2006 University of Toronto
%
% Permission is herby granted, without payment, to copy, modify, display
% and distribute this software and its documentation, if any, for any
% purpose, provided that the above copyright notices and the following
% three paragraphs appear in all copies of this software.  Use of this
% software constitutes acceptance of these terms and conditions.
%
% IN NO EVENT SHALL U of T BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
% SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF
% THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF U of T HAS BEEN ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%
% U of T SPECIFICALLY DISCLAIMS ANY EXPRESS OR IMPLIED WARRANTIES INCLUDING,
% BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
% FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.
%
% THIS SOFTWARE IS PROVIDED "AS IS," U of T HAS NO OBLIGATION TO PROVIDE
% MAINTENANCE, SUPPORT, UPDATE, ENHANCEMENTS, OR MODIFICATIONS.

