function varargout = MPI_Reduce(dest, tag, comm, func, varargin)
% MPI_Reduce - combines values from all processes to a single value
% 
% [var1, var2, ...] = MPI_Reduce(dest, tag, comm, func, var1, var2, ...)
%
% 'func' must be a handle to a binary function.  Choices for 'func'
% include @MPI_SUM, @MPI_PROD, @MPI_MIN, and @MPI_MAX.
%
% WARNING: for any process which as rank ~= dest, the return values
% of this function are garbage!  (Right now, for these processses, the
% input is simply copied to the output, but don't please don't rely on
% this behaviour.)

%----------------------------------------------------
% Check the input
%----------------------------------------------------

if nargin < 5
    error('too few input arguments');
end

if numel(varargin) ~= nargout
    warning(['number of input variables does not match number of '...
        'return values']);
end


%----------------------------------------------------
% Main function
%----------------------------------------------------

% Get processor rank.
my_rank = MPI_Comm_rank(comm);
comm_size = MPI_Comm_size(comm);

% if the comm has size 1, return without doing anything
if comm_size == 1
    varargout = varargin;
    return
end

result = varargin;
received = cell(1,numel(varargin));

% receive results from my right child
source = right_child(my_rank, dest);
if source < comm_size
    [received{:}] = MPI_Recv(source, tag, comm);
    for jj = 1:length(result)
        result{jj} = func(result{jj}, received{jj});
    end
end

% receive results from my left child
source = left_child(my_rank, dest);
if source < comm_size
    [received{:}] = MPI_Recv(source, tag, comm);
    for jj = 1:length(result)
        result{jj} = func(result{jj}, received{jj});
    end
end

% send the result to my parent (if I'm not the final destination)
if my_rank ~= dest
    MPI_Send(parent(my_rank, dest), tag, comm, result{:});
end

varargout = result;

end % end MPI_Reduce

%----------------------------------------------------
% Helper functions
%----------------------------------------------------

% Tree structure (assuming source == 0): 
%    root = 0, 
%    left child of node i = 2*i+1, 
%    right child of node i = 2*i+2.
%    parent(i) = ceil(ii/2)-1.
% When the source is not 0, we simply swap nodes '0' and 'source' in
% the tree, so 'source' is at the root and '0' is in 'source's
% position.  Otherwise the tree structure remains the same.
%
% The following funcitons simply compute the above relationsihps.

function p = parent(my_rank, source)
    my_rank = swap(my_rank, source);
    p = ceil(my_rank/2)-1;
    p = swap(p, source);
end

function l = left_child(my_rank, source)
    my_rank = swap(my_rank, source);    
    l = 2*my_rank+1;
    l = swap(l, source);
end

function r = right_child(my_rank, source)
    my_rank = swap(my_rank, source);    
    r = 2*my_rank+2;
    r = swap(r, source);
end

function x = swap(x,source)
    if x == source
        x = 0;
    elseif x == 0
        x = source;
    end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% MatlabMPI
% David Ross, Oct. 2006
% University of Toronto
% dross /at/ cs.toronto.edu
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright 2006 University of Toronto
%
% Permission is herby granted, without payment, to copy, modify, display
% and distribute this software and its documentation, if any, for any
% purpose, provided that the above copyright notices and the following
% three paragraphs appear in all copies of this software.  Use of this
% software constitutes acceptance of these terms and conditions.
%
% IN NO EVENT SHALL U of T BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
% SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF
% THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF U of T HAS BEEN ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%
% U of T SPECIFICALLY DISCLAIMS ANY EXPRESS OR IMPLIED WARRANTIES INCLUDING,
% BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
% FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.
%
% THIS SOFTWARE IS PROVIDED "AS IS," U of T HAS NO OBLIGATION TO PROVIDE
% MAINTENANCE, SUPPORT, UPDATE, ENHANCEMENTS, OR MODIFICATIONS.

